"Opening a czi file with RGB colors and read_mosaic function."

from position_to_tile import tiles_to_open
import matplotlib.pyplot as plt
import numpy as np
import pathlib
import aicspylibczi
from wavelength_to_color import get_the_wavelength


def norm_by(x, min_, max_):
    norms = np.percentile(x, [min_, max_])
    i2 = np.clip((x - norms[0]) / (norms[1] - norms[0]), 0, 1)
    return i2


def edges(image, x, y):
    """
    Add zeros to the numpy array if it is an egde array.

    Parameters
    ---------
    image : numpy.array
        The tile recolored.
    x : int
        XSize of a tile created by the microscope.
    y : int
        YSize of a tile created by the microscope.

    Return
    ---------
    newim : numpy.array
        The numpy.array with (x,y) size,
        filled with 0 if image is smaller than newim.

    """

    if (image.shape != (x, y, 3)):
        N = np.zeros((x, y, 3))
        (a, b, c) = image.shape
        N[0:a, 0:b] = image
        return (N)
    return(image)


def get_array(file, x0, x1, y0, y1, level):
    """
    Give the numpy array wanted with the read_mosaic function.
    With this function, you are able to choose the pyramid level.

    Parameters
    ---------
    file : str
       The path to your czi file.
    x0 : positive int
        The number of pixel on x axis at the beggining of the image wanted.
    x1 : positive int
        The number of pixel on x axis at the end of the image wanted.
    y0 : positive int
        The number of pixel on y axis at the beggining of the image wanted.
    y1 : positive int
        The number of pixel on y axis at the end of the image wanted.
    level : positive int
        The level of the pyramid wanted. Between 0 and 1.

    Example
    ---------
    file = '/Users/louisonrobach/Documents/renom.czi'

    Return
    ---------
    An array corresponding to the image wanted.
    """

    mosaic_file = pathlib.Path(file)
    czi = aicspylibczi.CziFile(mosaic_file)

    # We supposed the dimensions of each file to open is : 'HSCMYX'
    t = czi.size
    c = t[2]
    x = t[5]
    y = t[4]

    L = get_the_wavelength(file, c)

    mosaic_data = czi.read_mosaic(
        (x0, y0, x1-x0, y1-y0), scale_factor=level, C=0
    )
    mosaic_data = mosaic_data[0, :, :]
    c1 = (norm_by(mosaic_data, 50, 99.8) * 255).astype(np.uint8)
    c2 = (norm_by(mosaic_data, 50, 99.8) * 255).astype(np.uint8)
    c3 = (norm_by(mosaic_data, 0, 100) * 255).astype(np.uint8)
    rgb = np.stack((c1, c2, c3), axis=2)

# We recolor
    im_shape = np.array(rgb.shape)
    color_transform = np.array(L[0]).T
    im_reshape = rgb.reshape([np.prod(im_shape[0:2]), im_shape[2]]).T
    im_recolored = np.matmul(color_transform.T, im_reshape).T
    im_shape[2] = 3
    rgb = im_recolored.reshape(im_shape)
    mosaic_data = np.clip(rgb, 0, 255)
    mosaic_dataf = mosaic_data.astype(float) / 255.

    (x, y, v) = mosaic_dataf.shape

    image = np.zeros((x, y, 3), dtype=float)

    image += mosaic_dataf

    for k in range(1, c):
        mosaic_data = czi.read_mosaic(
            (x0, y0, x1-x0, y1-y0), scale_factor=level, C=k
        )
        mosaic_data = mosaic_data[0, :, :]
        c1 = (norm_by(mosaic_data, 50, 99.8) * 255).astype(np.uint8)
        c2 = (norm_by(mosaic_data, 50, 99.8) * 255).astype(np.uint8)
        c3 = (norm_by(mosaic_data, 0, 100) * 255).astype(np.uint8)
        rgb = np.stack((c1, c2, c3), axis=2)

    # We recolor
        im_shape = np.array(rgb.shape)
        color_transform = np.array(L[k]).T
        im_reshape = rgb.reshape([np.prod(im_shape[0:2]), im_shape[2]]).T
        im_recolored = np.matmul(color_transform.T, im_reshape).T
        im_shape[2] = 3
        rgb = im_recolored.reshape(im_shape)
        mosaic_data = np.clip(rgb, 0, 255)
        mosaic_dataf = mosaic_data.astype(float) / 255.

        mosaic_dataf = edges(mosaic_dataf, x, y)

        image += mosaic_dataf

    image /= 3.
    return(image)



"Give the RGB color given by the wavelength."

import xml.etree.ElementTree as ET
from xml.etree.ElementTree import ElementTree
import pathlib
import aicspylibczi


def WavelengthToColor(wavelength):
    """
    Give the RGB color corresponding to the wavelength.

    Parameter
    ---------
    wavelength : int
        The wavelength given in nanometer.

    Return
    ---------
    [R,G,B] : List
        The list of the RGB color, with R, G and B in [0,1].

    """
    if (wavelength == 0):
        red = 1.0  # white light !!!
        green = 1.0  # white light !!!
        blue = 1.0  # white light !!!

    elif ((wavelength >= 1) and (wavelength < 380)):
        red = 0.3
        green = 0.0
        blue = 0.3

    elif ((wavelength >= 380) and (wavelength < 440)):
        red = -(wavelength - 440) / (440 - 380)
        green = 0.0
        blue = 1.0

    elif ((wavelength >= 440) and (wavelength < 490)):
        red = 0.0
        green = (wavelength - 440) / (490 - 440)
        blue = 1.0

    elif ((wavelength >= 490) and (wavelength < 510)):
        red = 0.0
        green = 1.0
        blue = -(wavelength - 510) / (510 - 490)

    elif ((wavelength >= 510) and (wavelength < 540)):
        red = (wavelength - 510) / (540 - 510)
        green = 1.0
        blue = 0.0

    elif ((wavelength >= 540) and (wavelength < 620)):
        red = 1.0
        green = -(wavelength - 620) / (620 - 540)
        blue = 0.0

    elif ((wavelength >= 620) and (wavelength <= 680)):
        red = 1.0
        green = 0.0
        blue = (wavelength - 620) / (680 - 620)

    elif ((wavelength > 680) and (wavelength < 3000)):
        red = 0.3
        green = 0.0
        blue = 0.3

    else:
        red = 0.3
        green = 0.0
        blue = 0.3
    return([red, 0, 0], [0, green, 0], [0, 0, blue])


def get_the_wavelength(file, c):
    """
    Give the RGB color of each channel.

    Parameter
    ---------
    file : str
       The path to your czi file.
    c : int
        Number of channels present in the file.

    Example
    ---------
    file = '/Users/louisonrobach/Documents/renom.czi'

    Return
    ---------
    L : List
        The list composed of the RGB colors of each channels.

    """
    mosaic_file = pathlib.Path(file)
    czi = aicspylibczi.CziFile(mosaic_file)
    root = czi.meta
    superroot = ElementTree(root)
    root = superroot.getroot()

    L = []

    for k in range(0, c):
        for child in root[0][4][3][12][0][k]:
            if child.tag == "EmissionWavelength":
                a = float(child.text)
                L.append(WavelengthToColor(a))

    return(L)



"Opening a czi file without RGB colors and read_image function."

from position_to_tile import tiles_to_open
import matplotlib.pyplot as plt
import numpy as np
import pathlib
import aicspylibczi


def creates_numpy_array(file, n):
    """
    Open the n tile and creates a numpy array.

    Parameters
    ---------
    file : str
       The path to your czi file
    n : int

    Example
    ---------
    file = '/Users/louisonrobach/Documents/renom.czi'

    Return
    ---------
    M : np.array
        A numpy array that represents the colorization
        in RGB colors of the n tile.

    """
    mosaic_file = pathlib.Path(file)
    czi = aicspylibczi.CziFile(mosaic_file)

    # We supposed the dimensions of each file to open is : 'HSCMYX'
    t = czi.size
    c = t[2]
    x = t[5]
    y = t[4]

    image = np.zeros((x, y, 3), dtype=float)

    for k in range(0, c):
        tuple_mosaic_data = czi.read_image(M=n, C=k, S=0, H=0)
        mosaic_data = tuple_mosaic_data[0]  # We want only the numpy.darray
        mosaic_data = mosaic_data[0, 0, 0, 0, :, :]
        image += mosaic_data

    image /= 3.

    return(image)


def final_array_2(file, x0, x1, y0, y1):
    """
    Create the final numpy array that contains all tiles colored in RGB.

    Parameters
    ---------
    file : str
       The path to your czi file
    x0 : positive int
        The number of pixel on x axis at the beggining of the image wanted.
    x1 : positive int
        The number of pixel on x axis at the end of the image wanted.
    y0 : positive int
        The number of pixel on y axis at the beggining of the image wanted.
    y1 : positive int
        The number of pixel on y axis at the end of the image wanted.

    Example
    ---------
    file = '/Users/louisonrobach/Documents/renom.czi'

    Return
    ---------
    N : np.array
        A numpy array that contains every tiles colored in RGB.

    """

    mosaic_file = pathlib.Path(file)
    czi = aicspylibczi.CziFile(mosaic_file)
    x = czi.size[5]
    y = czi.size[4]

    L = tiles_to_open(file, x0, x1, y0, y1)

    k = 0
    while (L[k+1] == L[k]+1 and k < len(L)-2):
        k = k+1
    a = k+1

    M = []
    N = np.zeros((int(len(L)/a)*y, a*x, 3))

    for i in range(0, int(len(L)/a)):
        for j in range(0, a):
            image = creates_numpy_array(file, L[a*i+j])
            M.append(image)

    d = 0
    for i in range(0, int(len(L)/a)*y, y):
        for j in range(0, a*x, x):
            N[i:i+y, j:j+x] = M[d]
            d += 1

    return(N)


def creates_non_RGB_array(file, x0, x1, y0, y1):
    """
    Create the numpy array containing the non RGB image.

    Parameters
    ---------
    file : str
       The path to your czi file.
    x0 : positive int
        The number of pixel on x axis at the beggining of the image wanted.
    x1 : positive int
        The number of pixel on x axis at the end of the image wanted.
    y0 : positive int
        The number of pixel on y axis at the beggining of the image wanted.
    y1 : positive int
        The number of pixel on y axis at the end of the image wanted.

    Example
    ---------
    file = '/Users/louisonrobach/Documents/renom.czi'

    """

    N = final_array_2(file, x0, x1, y0, y1)



"Opening a czi file with RGB colors and read_image function."

from position_to_tile import tiles_to_open
import matplotlib.pyplot as plt
import numpy as np
import pathlib
import aicspylibczi
from wavelength_to_color import get_the_wavelength


def norm_by(x, min_, max_):
    norms = np.percentile(x, [min_, max_])
    i2 = np.clip((x - norms[0]) / (norms[1] - norms[0]), 0, 1)
    return i2


def recolor_tile(file, n):
    """
    Open the n tile and creates a numpy array colored in RGB.

    Parameters
    ---------
    file : str
       The path to your czi file.
    n : int

    Example
    ---------
    file = '/Users/louisonrobach/Documents/renom.czi'

    Return
    ---------
    M : np.array
        A numpy array that represents the colorization
         in RGB colors of the n tile.

    """
    mosaic_file = pathlib.Path(file)
    czi = aicspylibczi.CziFile(mosaic_file)

    # We supposed the dimensions of each file to open is : 'HSCMYX'

    t = czi.size
    c = t[2]
    x = t[5]
    y = t[4]

    L = get_the_wavelength(file, c)

    image = np.zeros((x, y, 3), dtype=float)

    for k in range(0, c):
        tuple_mosaic_data = czi.read_image(M=n, C=k, S=0, H=0)
        mosaic_data = tuple_mosaic_data[0]  # We want only the numpy.darray
        mosaic_data = mosaic_data[0, 0, 0, 0, :, :]
        c1 = (norm_by(mosaic_data, 50, 99.8) * 255).astype(np.uint8)
        c2 = (norm_by(mosaic_data, 50, 99.8) * 255).astype(np.uint8)
        c3 = (norm_by(mosaic_data, 0, 100) * 255).astype(np.uint8)
        rgb = np.stack((c1, c2, c3), axis=2)

    # On recolore
        im_shape = np.array(rgb.shape)
        color_transform = np.array(L[k]).T
        im_reshape = rgb.reshape([np.prod(im_shape[0:2]), im_shape[2]]).T
        im_recolored = np.matmul(color_transform.T, im_reshape).T
        im_shape[2] = 3
        rgb = im_recolored.reshape(im_shape)
        mosaic_data = np.clip(rgb, 0, 255)
        mosaic_dataf = mosaic_data.astype(float) / 255.

        mosaic_dataf = edges(mosaic_dataf, x, y)

        image += mosaic_dataf

# We are making an average color to realize a correct superimposition of colors
    image /= 3.

    return(image)


def edges(image, x, y):
    """
    Add zeros to the numpy array if it is an egde array.

    Parameters
    ---------
    image : numpy.array
        The tile recolored.
    x : int
        XSize of a tile created by the microscope.
    y : int
        YSize of a tile created by the microscope.

    Return
    ---------
    newim : numpy.array
        The numpy.array with (x,y) size,
        filled with 0 if image is smaller than newim.

    """

    if (image.shape != (x, y, 3)):
        N = np.zeros((x, y, 3))
        (a, b, c) = image.shape
        N[0:a, 0:b] = image
        return (N)
    return(image)


def final_array(file, x0, x1, y0, y1):
    """
    Create the final numpy array that contains all tiles colored in RGB.

    Parameters
    ---------
    file : str
       The path to your czi file.
    x0 : positive int
        The number of pixel on x axis at the beggining of the image wanted.
    x1 : positive int
        The number of pixel on x axis at the end of the image wanted.
    y0 : positive int
        The number of pixel on y axis at the beggining of the image wanted.
    y1 : positive int
        The number of pixel on y axis at the end of the image wanted.

    Example
    ---------
    file = '/Users/louisonrobach/Documents/renom.czi'

    Return
    ---------
    N : np.array
        A numpy array that contains every tiles colored in RGB.

    """

    mosaic_file = pathlib.Path(file)
    czi = aicspylibczi.CziFile(mosaic_file)
    x = czi.size[5]
    y = czi.size[4]

    U = tiles_to_open(file, x0, x1, y0, y1)

    k = 0
    while (U[k+1] == U[k]+1 and k < len(U)-2):
        k = k+1
    a = k+1

    M = []
    N = np.zeros((int(len(U)/a)*y, int(a*x), 3))

    for i in range(0, int(len(U)/a)):
        for j in range(0, a):
            image = recolor_tile(file, U[a*i+j])
            M.append(image)

    d = 0
    for i in range(0, int(len(U)/a)*y, y):
        for j in range(0, a*x, x):
            N[i:i+y, j:j+x] = M[d]
            d += 1

    return(N)


def size_from_tiles(file, x0, x1, y0, y1):
    """
    Give the distance from the edge of the tiles where the image is.

    Parameters
    ---------
    file : str
       The path to your czi file.
    x0 : positive int
        The number of pixel on x axis at the beggining of the image wanted.
    x1 : positive int
        The number of pixel on x axis at the end of the image wanted.
    y0 : positive int
        The number of pixel on y axis at the beggining of the image wanted.
    y1 : positive int
        The number of pixel on y axis at the end of the image wanted.

    Example
    ---------
    file = '/Users/louisonrobach/Documents/renom.czi'

    Return
    ---------
    [x2, x3, y2, y3] : List
        The list containing the x and y position of the image
        in the surrounding tiles.

    """
    mosaic_file = pathlib.Path(file)
    czi = aicspylibczi.CziFile(mosaic_file)

    # We supposed the dimensions of each file to open is : 'HSCMYX'
    t = czi.size
    pix = t[5]
    piy = t[4]

    x2 = x0 % pix
    y2 = y0 % piy
    x3 = x2 + (x1-x0)
    y3 = y2 + (y1-y0)
    return([x2, x3, y2, y3])


def display_image(file, x0, x1, y0, y1):
    """
    Display the RGB image with the position in file wanted.

    Parameters
    ---------
    file : str
       The path to your czi file.
    x0 : positive int
        The number of pixel on x axis at the beggining of the image wanted.
    x1 : positive int
        The number of pixel on x axis at the end of the image wanted.
    y0 : positive int
        The number of pixel on y axis at the beggining of the image wanted.
    y1 : positive int
        The number of pixel on y axis at the end of the image wanted.

    Example
    ---------
    file = '/Users/louisonrobach/Documents/renom.czi'

    Return
    ---------
    An image created by matplotlib corresponding to the image wanted.
    """

    N = final_array(file, x0, x1, y0, y1)

    [x2, x3, y2, y3] = size_from_tiles(file, x0, x1, y0, y1)

    N = N[y2:y3, x2:x3]

    plt.imshow(N)
    plt.axis('on')
    plt.show()



from opening_with_pyramid_level import get_array
import pathlib
import aicspylibczi
from position_to_tile import size


class FluOpenSlide():
    """
    The FluOpenSlide object try to behave like the OpenSlide one.
    The present functions have similar interests than OpenSlide ones,
    and so similar names.
    """

    def __init__(self, file):
        """
        Open the .czi file.

        Parameter
        ---------
        file : str
           The path to your czi file.

        """
        self.path = file

        mosaic_file = pathlib.Path(self.path)
        czi = aicspylibczi.CziFile(mosaic_file)
        self.file = czi
        self.fluo = True

    def data(self):
        """
        Give information about the file. Returns a list : [[SizeX, SizeY],
        number of tiles, [tileSizeX, tileSizeY]].

        """
        t = self.file.size
        return([size(self.path), t[3], [t[5], t[4]]])

    def level_count(self):
        """
        Give the number of levels in the image.
        This number is not real, because there is an infinity of levels
        accessible with the scale_factor.

        """
        return(10)

    def levels(self):
        """
        Give a list containing the different levels of the image.

        """
        return sorted([1./(2**k) for k in range(10)])

    def level_dimensions(self, location, size):
        """
        A list of (width, height) tuples, one for each level of the image.

        """
        L = []
        x, y = location
        w, h = size
        for k in self.levels():
            mosaic_data = self.file.read_mosaic(
                (x, y, w, h), scale_factor=k, C=0
            )
            mosaic_data = mosaic_data[0, :, :]
            L.append(mosaic_data.shape)
        return(L)

    def level_downsamples(self):
        """
        A list of downsample factors for each level of the slide.
        level_downsamples[k] is the downsample factor of level k.

        """
        return sorted(2**k for k in range(10))

    def read_region(self, location, level, size):
        """
        Display the image with the same arguments than used with openslide.

        Parameters
        ---------
        location = (x, y) : tuple
            tuple giving the top left pixel of the image wanted.
        level : int
            the level number.
        size = (width, height) : tuple
            tuple giving the size of the image.

        """

        (x, y) = location
        (w, h) = size

        return (get_array(self.path, x, x+w, y, y+h, level/(2**level)))



""" By giving position, width, height in pixels,
you have access to the tiles needed to open your flurescence image."""

import pathlib
import aicspylibczi
from xml.etree.ElementTree import ElementTree
import xml.etree.ElementTree as ET
from typing import Tuple, List


def size(file) -> Tuple[int, int]:
    """
    Find the dimension of the entire czi image.

    Parameter
    ---------
    file : str
        The name of your czi file

    Example
    ---------
    file = '/Users/louisonrobach/Documents/renom.czi'

    Return
    ---------
    Tuple [SizeX, SizeY]
        The size, in pixel, of your entire czi image.

    """
    L = []
    mosaic_file = pathlib.Path(file)
    czi = aicspylibczi.CziFile(mosaic_file)

    root = czi.meta
    superroot = ElementTree(root)
    root = superroot.getroot()

    for child in root[0][4][3]:
        if child.tag == "SizeX" or child.tag == "SizeY":
            L.append(float(child.text))

    return(L)


def tiles_to_open(file, x0, x1, y0, y1) -> List:
    """
    Create a list with the numbers of tiles to open.

    Parameters
    ---------
    file : str
       The name of your czi file
    x0 : positive int
        The number of pixel on x axis at the beggining of the image wanted.
    x1 : positive int
        The number of pixel on x axis at the end of the image wanted.
    y0 : positive int
        The number of pixel on y axis at the beggining of the image wanted.
    y1 : positive int
        The number of pixel on y axis at the end of the image wanted.

    Return
    ---------
    L : List
        A list of numbers representing the tiles
        to open to get the image wanted.

    """

    L = []

    mosaic_file = pathlib.Path(file)
    czi = aicspylibczi.CziFile(mosaic_file)

    # We supposed the dimensions of each file to open is : 'HSCMYX'
    t = czi.size
    pix = t[5]
    piy = t[4]
    [SizeX, SizeY] = size(file)
    column0 = x0//pix
    column1 = x1//pix
    line0 = y0//piy
    line1 = y1//piy
    nbXtile = SizeX//pix + 1

    for j in range(line0, line1 + 1):
        a = int(column0 + j * nbXtile)
        b = int(column1 + j * nbXtile)

        for k in range(a, b+1):
            L.append(k)

    return(L)



"""Setup script to install fluopenslide."""

from setuptools import setup, find_packages
# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="fluopenslide1",
    version="0.0.1",
    description="Opening fluorescence .czi files",
    author="Louison Robach",
    author_email="louison.robach@gmail.com",
    packages=find_packages(),
    zip_safe=False,
    install_requires=[
        "numpy",
        "aicspylibczi",
        "matplotlib",
        "pathlib",
        "aicsimageio"
    ],
    include_package_data=True,
    long_description=long_description,
    long_description_content_type='text/markdown'
)



from fluopenslide import FluOpenSlide
from openslide import OpenSlide
import os
from os.path import basename


L = [
    '.mrxs', '.svs', '.tif', '.ndpi', '.vms', '.vmu',
    '.scn', '.tiff', '.svslide', '.bif'
]


class OpenslideObject(Exception):
    """
    The object created can be opened with OpenSlide.
    """
    pass


class FluOpenSlideObject(Exception):
    """
    The object created can be opened with FluOpenSlide.
    """
    pass


class Unknown(Exception):
    """
    The object can not be opened.
    """
    pass


class choice(FluOpenSlide, OpenSlide):
    """
    Decide witch program use between OpenSlide and FluOpenSlide.

    """
    def __init__(self, file, choice=None):
        """
        Parameters
        ---------
        file : str
           The path to your czi file.
        choice : str
            Have to be "FluOpenSlide" or "Openslide".
            If there is no choice, focus on the file extension.
        """
        self.name, self.extension = os.path.splitext(file)

        if (choice == "FluOpenSlide" or choice == "fluopenslide"):
            FluOpenSlide.__init__(self, file)
            self = FluOpenSlide(file)
            raise FluOpenSlideObject(
                'Le fichier est à présent un objet FluOpenSlide.'
            )
        if (choice == "Openslide" or choice == "openslide"):
            OpenSlide.__init__(self, file)
            self = Openslide(file)
            raise OpenslideObject(
                'Le fichier est à présent un objet OpenSlide.'
            )
        if choice is None:
            print(
                "Aucune ouverture spécifique du fichier"
                "n'a été demandée.\n"
                "Par conséquent, "
                "nous allons nous baser sur l'extension du fichier."
            )
            if (self.extension == '.czi'):
                FluOpenSlide.__init__(self, file)
                self = FluOpenSlide(file)
                raise FluOpenSlideObject("L'objet est un FluOpenSlide.")
            if (self.extension in L):
                OpenSlide.__init__(self, file)
                self = OpenSlide(file)
                raise OpenslideObject("L'objet est un OpenSlide.")
            raise Unknown("Le type de fichier n'est pas ouvrable.")
        raise Unknown(
            "Aucune syntaxe n'a été reconnue."
            "Merci de lire la documentation avant de rééssayer."
        )
